@@ -1,4 +1,4 @@
-/* $Id: dcc7.c 610 2008-05-31 22:18:03Z wojtekka $ */
+/* $Id: dcc7.c 883 2009-12-24 14:24:41Z wojtekka $ */
 
 /*
  *  (C) Copyright 2001-2008 Wojtek Kaniewski <wojtekka@irc.pl>
@@ -396,9 +396,7 @@ static struct gg_dcc7 *gg_dcc7_send_file_common(struct gg_session *sess, uin_t r
 	return dcc;
 
 fail:
-	if (dcc)
-		free(dcc);
-
+	free(dcc);
 	return NULL;
 }
 
@@ -472,9 +470,7 @@ fail:
 		errno = errsv;
 	}
 
-	if (dcc)
-		free(dcc);
-
+	free(dcc);
 	return NULL;
 }
 
@@ -730,6 +726,7 @@ int gg_dcc7_handle_info(struct gg_session *sess, struct gg_event *e, void *paylo
 		e->event.dcc7_accept.remote_port = dcc->remote_port;
 	} else {
 		e->type = GG_EVENT_DCC7_PENDING;
+		e->event.dcc7_pending.dcc7 = dcc;
 	}
 
 	if (gg_dcc7_connect(sess, dcc) == -1) {
@@ -1003,6 +1000,7 @@ struct gg_event *gg_dcc7_watch_fd(struct gg_dcc7 *dcc)
 
 				if (gg_dcc7_reverse_connect(dcc) != -1) {
 					e->type = GG_EVENT_DCC7_PENDING;
+					e->event.dcc7_pending.dcc7 = dcc;
 				} else {
 					e->type = GG_EVENT_DCC7_ERROR;
 					e->event.dcc_error = GG_ERROR_DCC7_NET;
